from pwn import *
import subprocess
import json

#Cloned from here https://github.com/icemonster/symbolic_mersenne_cracker/
from symbolic_mersenne_cracker.main import Untwister

from math import pi, acos, sqrt
import time

context.log_level = 'CRITICAL' #Just shut up pwntools


### ----------- UTILS ----------------------------------
NUM_QUBITS = 30
XGATE = 1
ZGATE = 3
HADAMARD = 4
RYGATE = 7
CNOT = 9
SWAP = 10
UNITARY = 12
MEASURE = 13

def unitary(c, matrix, qubit):
    return interact(c, {'option': 'op_on_med', 'operation': UNITARY, 'qubit': qubit, 'params': matrix})

def rygate(c, theta, bit, qubit=0):
    return interact(c, {'option': 'op_on_med', 'operation': RYGATE, 'qubit': qubit, 'params':[theta, bit]})

def xgate(c, bit, qubit=0):
    return interact(c, {'option': 'op_on_med', 'operation': XGATE, 'qubit': qubit, 'params':[bit]})

def zgate(c, bit, qubit=0):
    return interact(c, {'option': 'op_on_med', 'operation': ZGATE, 'qubit': qubit, 'params':[bit]})

def cnot(c, control_bit, other_bit, qubit=0):
    return interact(c, {'option': 'op_on_med', 'operation': CNOT, 'qubit': qubit, 'params':[control_bit, other_bit]})

def hadamard(c, bit, qubit=0):
    return interact(c, {'option': 'op_on_med', 'operation': HADAMARD, 'qubit': qubit, 'params':[bit]})

def swap(c, bit1, bit2, qubit=0):
    return interact(c, {'option': 'op_on_med', 'operation': SWAP, 'qubit': qubit, 'params': [bit1, bit2]})

def measure(c, bit, qubit):
    return interact(c, {'option': 'op_on_med', 'operation': MEASURE, 'qubit': qubit, 'params': [bit]})

def buy_medallion(c):
    out = interact(c, {'option': 'money_for_med'})
    return out['msg'].split(' ')[3]  #Parses medallion id and returns it

def sell_medallion(c, med_id):
    out = interact(c, {'option': 'med_for_money', 'med_id': med_id})
    return 'error' not in out  #Returns whether it was successful

# ----------- END UTILS --------------------------------


def get_res(old_bits):
    '''
        Just returns bits and bases as bitstrings
    '''
    bits = ''
    bases = ''
    for i in old_bits:
        if i == '0':
            bits += '0'
            bases += '0'
        elif i == '1':
            bits += '1'
            bases += '0'
        elif i == '+':
            bits += '0'
            bases += '1'
        elif i == '-':
            bits += '1'
            bases += '1'
    return bits[::-1], bases[::-1]


def prepare_state_cloning(c):
    for qubit in range(NUM_QUBITS):
        #theta1 = acos(1/sqrt(5))/2
        #theta2 = acos(sqrt(5)/3)/2
        #theta3 = acos(2/sqrt(5))/2
        #rygate(c, theta1*2, 1, qubit)
        #cnot(c, 1, 2, qubit)
        #rygate(c, theta2*2, 2, qubit)
        #cnot(c, 2, 1, qubit)
        #rygate(c, theta3*2, 1, qubit)
        #cnot(c, 0, 1, qubit)
        #cnot(c, 0, 2, qubit)
        #cnot(c, 1, 0, qubit)
        #cnot(c, 2, 0, qubit)
        #Sending a single request is faster
        unitary(c, [[0.816496580927726, 0, -0.40824829046386296, 0, -0.4082482904638631, 0, 0, 0], [0, 0.4082482904638631, 0, 1.1714520289030084e-17, 0, 0.816496580927726, 0, -0.408248290463863], [0, 0.40824829046386296, 0, 0.816496580927726, 0, -4.767256628539764e-18, 0, 0.4082482904638631], [1.1714520289030084e-17, 0, -0.4082482904638631, 0, 0.408248290463863, 0, 0.816496580927726, 0], [0, 1.1714520289030084e-17, 0, -0.4082482904638631, 0, 0.408248290463863, 0, 0.816496580927726], [0.40824829046386296, 0, 0.816496580927726, 0, -4.767256628539764e-18, 0, 0.4082482904638631, 0], [0.4082482904638631, 0, 1.1714520289030084e-17, 0, 0.816496580927726, 0, -0.408248290463863, 0], [0, 0.816496580927726, 0, -0.40824829046386296, 0, -0.4082482904638631, 0, 0]], qubit)


def restore_cloned_state(c):
    for qubit in range(NUM_QUBITS):
        swap(c, 0, 1, qubit)

def state_reconstruct(c, bases, bits):
    ''' Copied from challenge '''
    for qubit in range(NUM_QUBITS):
        measure(c, 0, qubit=qubit)
        if bases % 2:
            if bits % 2:
                # This gives |->
                xgate(c, 0, qubit=qubit)
                hadamard(c, 0, qubit=qubit)
            else:
                # This gives |+>
                hadamard(c, 0, qubit=qubit)
        else:
            if bits % 2:
                # This gives |1>
                xgate(c, 0, qubit=qubit)
            else:
                # This gives |0>, cuz why not
                xgate(c, 0, qubit=qubit)
                xgate(c, 0, qubit=qubit)

        bases //= 2
        bits //= 2


def interact(server, data):
    server.sendline(json.dumps(data))
    return json.loads(server.recvline().strip())

def exploit():
    num_tries = 0

    #This loop tries to sell a medallion twice
    #   by cloning the quantum state, sell the medallion,
    #   restore the cloned state and sell it again
    while 1:
        num_tries += 1
        print('Attempt:', num_tries)

        #Connect and solve POW
        p = remote('fake.medallion.pwni.ng', 15213)
        #p = remote('localhost',15213)
        resource = p.recvline().strip()
        out = subprocess.check_output(['hashcash','-b','21','-m',resource])
        out = out.split(b' ')[-1]
        p.sendline(out)
        _ = p.recvuntil(b'evious result>}\n')
        _ = p.clean()


        med_id = buy_medallion(p)

        prepare_state_cloning(p)

        if sell_medallion(p, med_id):
            restore_cloned_state(p)

            #If we can sell the same medallion twice we are done
            if sell_medallion(p, med_id):
                break

        p.close()

    print('We managed to sell the same medallion twice')
    med_id = buy_medallion(p)

    ut = Untwister()

    for i in range(624):
        print('Cloning Random state:', i,'/',624)

        out = interact(p, {'option': 'med_for_med', 'med_id': med_id})
        med_id = out['msg'].split(' ')[2]
        old_med = out['msg'].split(' ')[-6][:-1]
        bits, bases = get_res(old_med)

        ut.submit(bases + '?'*(32 - NUM_QUBITS)) #2 unknown bits because getrandbits(30) truncated 2 lsbs
        ut.submit(bits + '?'*(32 - NUM_QUBITS))

    #Obtain cloned random
    r = ut.get_random()

    bases = r.getrandbits(NUM_QUBITS) #current bases
    bits = r.getrandbits(NUM_QUBITS)  #current bits

    print('Getting the flag...')

    for _ in range(14):
        assert sell_medallion(p, med_id)
        state_reconstruct(p, bases, bits)

    print(interact(p, {'option': 'med_for_money', 'med_id': med_id})) #Get the flag

exploit()
